
#include <cstdint>
#include <cstdio>
#include <cassert>

#include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/irdb/STARSProgram.h"

#include "base/SMPProgram.h"

using namespace std;

#define SMP_DEBUG 1


// Set internal state to handle a 32-bit binary
void STARS_IRDB_Program_t::Set32BitBinary(void) {
	STARS_Program_t::SetBitwidth32();
	// this->STARS_ISA_dtyp = dt_dword;
	// this->STARS_MD_LAST_SAVED_REG_NUM = R_di;
	return;
}

// Set internal state to handle a 64-bit binary
void STARS_IRDB_Program_t::Set64BitBinary(void) {
	STARS_Program_t::SetBitwidth64();
	// this->STARS_ISA_dtyp = dt_qword;
	// this->STARS_MD_LAST_SAVED_REG_NUM = R_r15;
	return;
}

// Set flags, take actions based on code size.
void STARS_IRDB_Program_t::ReportTotalCodeSize(unsigned long long TotalCodeSize) {
	this->SetTotalCodeSize(TotalCodeSize);
	// this->SetReducedProcessingFlag(TotalCodeSize > STARS_CODESIZE_FULL_ANALYSIS_LIMIT);
	return;
}

bool STARS_IRDB_Program_t::OpenFiles(void) {
#if 0 // Need IRDB version of GetRootFileName()
	string ZSTLogFileName(global_STARS_program->GetRootFileName());
	string LogFileSuffix(".STARSlog");
	ZSTLogFileName += LogFileSuffix;
	this->SMPLogFile = SMP_fopen(ZSTLogFileName.c_str(), "w");
	if (NULL == SMPLogFile) {
		SMP_msg("ERROR: Cannot open STARS log file %s\n", ZSTLogFileName.c_str());
		SMP_msg("Redirecting to stderr.\n");
		this->SMPLogFile = stderr;
	}
#endif

	// Open all the common files via the base class.
	return STARS_Program_t::OpenFiles();
} // end of STARS_IRDB_Program_t::OpenFiles()

void STARS_IRDB_Program_t::CloseFiles(void) {
	// Close all the common files via the base class.
	STARS_Program_t::CloseFiles();
	return;
} // end of STARS_IRDB_Program_t::CloseFiles()

void STARS_IRDB_Program_t::InitData(void) {
	SMP_msg("INFO: Reach STARS_IDA_Program_t::InitData method.\n");

	// this->MDInitializeArgumentRegs();
	// this->MDInitializeCallerSavedRegs();

	// Init everything else through the base class.
	STARS_Program_t::InitData();

#if 0
	// Record the start and end addresses for all function entry
	//  chunks in the program.
	this->FuncBounds.reserve(10 + SMP_get_func_qty());
	for (std::size_t FuncIndex = 0; FuncIndex < SMP_get_func_qty(); ++FuncIndex) {
		STARS_Function_t *FuncInfo = SMP_getn_func(FuncIndex);
		SMP_bounds_t temp;
		temp.startEA = FuncInfo->get_startEA();
		temp.endEA = FuncInfo->get_endEA();
		FuncBounds.push_back(temp);
	}
#endif

	return;
} // end of STARS_IRDB_Program_t::InitData()

#if 0
void STARS_IRDB_Program_t::DetermineRootFileName(void) {
	SMP_msg("INFO: Reached DetermineRootFileName method.\n");
	// Get the root file name from IDA Pro, e.g. "foo" when analyzing "foo.exe"
	char TempRootName[STARS_MAXSTR];
	STARS_ssize_t FileLen = ::get_root_filename(TempRootName, sizeof(TempRootName)-1);
	string TempRootString(TempRootName);
	this->SetRootFileName(TempRootString);
	return;
}
#endif

#if 0
void STARS_IRDB_Program_t::MDInitializeCallerSavedRegs(void) {
	this->STARS_MDCallerSavedRegs.clear();
	bool x86_64_ISA_flag = false;
#ifdef __EA64__
	x86_64_ISA_flag = (this->GetSTARS_ISA_Bitwidth() == 64);
#endif
	if (!x86_64_ISA_flag) {
		// 32-bit x86 uses EAX, ECX, EDX as caller-saved.
		this->STARS_MDCallerSavedRegs.push_back(R_ax);
		this->STARS_MDCallerSavedRegs.push_back(R_cx);
		this->STARS_MDCallerSavedRegs.push_back(R_dx);
	}
	else {
		// 64-bit x86 uses EDI, ESI, EDX, ECX, R8 and R9
		//  in that order. After six arguments that fit into
		//  these regs, arguments are passed on the stack.
		// In addition, registers EAX, R10 and R11 are caller-saved
		//  but are not used to pass arguments.
		this->STARS_MDCallerSavedRegs.push_back(R_ax);
		this->STARS_MDCallerSavedRegs.push_back(R_cx);
		this->STARS_MDCallerSavedRegs.push_back(R_dx);
		this->STARS_MDCallerSavedRegs.push_back(R_si);
		this->STARS_MDCallerSavedRegs.push_back(R_di);
		this->STARS_MDCallerSavedRegs.push_back(R_r8);
		this->STARS_MDCallerSavedRegs.push_back(R_r9);
		this->STARS_MDCallerSavedRegs.push_back(R_r10);
		this->STARS_MDCallerSavedRegs.push_back(R_r11);
	}
	return;
} // end of STARS_IRDB_Program_t::MDInitializeCallerSavedRegs()
#endif

#if 0
void STARS_IRDB_Program_t::MDInitializeArgumentRegs(void) {
	bool x86_64_ISA_flag = false;
#ifdef __EA64__
	x86_64_ISA_flag = (this->GetSTARS_ISA_Bitwidth() == 64);
#endif
	if (x86_64_ISA_flag) {
		this->STARS_MDArgumentRegs.push_back(R_di);
		this->STARS_MDArgumentRegs.push_back(R_si);
		this->STARS_MDArgumentRegs.push_back(R_dx);
		this->STARS_MDArgumentRegs.push_back(R_cx);
		this->STARS_MDArgumentRegs.push_back(R_r8);
		this->STARS_MDArgumentRegs.push_back(R_r9);
	}
	else {
		this->STARS_MDArgumentRegs.clear();
	}
	return;
} // end of STARS_IRDB_Program_t::MDInitializeArgumentRegs()
#endif

// Does the instruction at InstAddr access the global data offset in GlobalAddr
//  using an index register?
bool MDIsIndexedAccess(STARS_ea_t InstAddr, STARS_ea_t GlobalAddr) {
	bool DebugFlag = (InstAddr == 0x80502d3);
#if SMP_DETECT_INDEXED_ACCESSES
	uint32_t LocFeatures;
	insn_t LocalCmd;
	SMPInstr TempInst(InstAddr);
	bool InstOK = TempInst.FillCmd();
	if (!InstOK)
#endif
		return false;

#if SMP_DETECT_INDEXED_ACCESSES
	for (int i = 0; i < UA_MAXOP; ++i) {
		STARSOpndTypePtr CurrOp = LocalCmd.Operands[i];
		if (CurrOp->IsStaticMemOp() || CurrOp->IsMemDisplacementOp()) {
			if (GlobalAddr == CurrOp.addr) {
				if (CurrOp->HasSIBByte()) {
					// GlobalAddr is referenced, and SIB byte is present, so we might have
					//  an indexed access to GlobalAddr.
					short IndexReg = sib_index(CurrOp);
					if (STARS_x86_R_sp != IndexReg) {
						// STARS_x86_R_sp is a SIB index dummy value; means no index register
						return true;
					}
				}
				else if (CurrOp->IsMemDisplacementOp()) { // index reg in reg field, not in SIB byte
					return true;
				}
				else if (DebugFlag) {
					SMP_msg("Failed to find index in operand: ");
					PrintOneOperand(CurrOp, 0, -1);
					SMP_msg("\n");
				}
			}
		}
	} // end for all operands
	return false;
#endif
} // end MDIsIndexedAccess()

#if 0
// Determine static global variable boundaries.
void STARS_IRDB_Program_t::InitStaticDataTable(SMPProgram *CurrProg) {
	char buf[STARS_MAXSTR];
	STARS_ea_t ea;
	flags_t ObjFlags;
	bool ReadOnlyFlag;
	size_t DummyNumber = 0; // suffix for SMP_dummy# global names

	// First, examine the data segments and collect info about static
	//   data, such as name/address/size.

	STARS_ea_t TempLowestGlobalVarAddress = STARS_BADADDR;
	STARS_ea_t TempHighestGlobalVarAddress = 0x00000000;
	STARS_ea_t TempLowestCodeAddress = STARS_BADADDR;
	STARS_ea_t TempHighestCodeAddress = 0x00000000;

	// Loop through all segments.
	STARS_ea_t RecentAddr = STARS_BADADDR;
#if IDA_SDK_VERSION < 600
	for (int SegIndex = 0; SegIndex < SMP_get_segm_qty(); ++SegIndex) {
		segment_t *seg = SMP_getnseg(SegIndex);
#else
	for (STARS_Segment_t *seg = SMP_get_first_seg(); NULL != seg; seg = SMP_get_next_seg(RecentAddr)) {
#endif
		char SegName[STARS_MAXSTR];
		RecentAddr = seg->get_startEA();
		STARS_ssize_t SegNameSize = SMP_get_segm_name(seg, SegName, sizeof(SegName) - 1);

		// We are only interested in the data segments of type
		// SEG_DATA, SEG_BSS and SEG_COMM.
#if SMP_DEBUG
		SMP_msg("Found segment of type <elided>" /*, seg->type */);
		if (SegNameSize > 0)
			SMP_msg(" SegName: %s", SegName);
		SMP_msg(" from %lx to %lx\n", (unsigned long)seg->get_startEA(), (unsigned long)seg->get_endEA());
		if (ReadOnlyFlag) {
			SMP_msg("Read-only segment.\n");
		}
#endif
		if ((seg->IsDataSegment()) || (seg->IsBSSSegment()) || (seg->IsCommonSegment())) {
			// Loop through each of the segments we are interested in,
			//  examining all data objects (effective addresses).
			ReadOnlyFlag = ((seg->IsReadableSegment()) && (!(seg->IsWriteableSegment())));
#if SMP_DEBUG
			SMP_msg("Starting data segment of type <elided>" /*, seg->type */);
			if (SegNameSize > 0)
				SMP_msg(" SegName: %s", SegName);
			SMP_msg(" from %lx to %lx\n", (unsigned long)seg->get_startEA(), (unsigned long)seg->get_endEA());
			if (ReadOnlyFlag) {
				SMP_msg("Read-only data segment.\n");
			}
#endif
			ea = seg->get_startEA();
			while (ea < seg->get_endEA()) {
				ObjFlags = SMP_get_flags_novalue(ea);
				// Only process head bytes of data objects, i.e. isData().
				if (isData(ObjFlags)) {
					// Compute the size of the data object.
					STARS_ea_t NextEA = SMP_next_head(ea, seg->get_endEA());
					if (NextEA == BADADDR) {
						NextEA = seg->get_endEA();
					}
					size_t ObjSize = (size_t) (NextEA - ea);
					if (TempLowestGlobalVarAddress > ea)
						TempLowestGlobalVarAddress = ea;
					if (TempHighestGlobalVarAddress < NextEA)
						TempHighestGlobalVarAddress = NextEA - 1;
					// Get the data object name using its address.
					char *TrueName = SMP_get_true_name(BADADDR, ea, buf, sizeof(buf));
					if (NULL == TrueName) {
						int count = SMP_snprintf(buf, sizeof(buf), "SMP_dummy%zu", DummyNumber);
						++DummyNumber;
						TrueName = buf;
					}

					// Record the name, address, size, and type info.
					struct GlobalVar VarTemp;
					VarTemp.addr = ea;
					VarTemp.size = ObjSize;
					VarTemp.ReadOnly = ReadOnlyFlag;
					VarTemp.IndexedAccess = false;
					VarTemp.flags = ObjFlags;
					SMP_strncpy(VarTemp.name, buf, sizeof(VarTemp.name) - 1);
					// Insert name and address into name-addr map if it is
					//  not an SMP_dummy0 variable.
					if (NULL != TrueName) {
						string bufstring(buf);
						pair<string, STARS_ea_t> TempPair(bufstring, ea);
						pair<map<string, STARS_ea_t>::iterator, bool> InsertResult;
						InsertResult = CurrProg->InsertGlobalNameAddrMapEntry(TempPair);
						if (!InsertResult.second) {
							SMP_msg("FATAL ERROR: Insertion into GlobalNameMap: %s at %llu\n", buf, ea);
						}
						assert(InsertResult.second);
					}
					VarTemp.FieldOffsets.clear();
#if SMP_COMPUTE_GLOBAL_GRANULARITY
					if (VarTemp.size < 1000000) { // don't waste time on monster objects
						this->ComputeGlobalFieldOffsets(VarTemp);
					}
#endif
					pair<STARS_ea_t, struct GlobalVar> TempItem(ea, VarTemp);
					CurrProg->InsertGlobalVarTableEntry(TempItem);

					// Check for code xrefs from the data.
					// Can have a table of pointers, so iterate through large data objects.
					STARS_ea_t TempAddr = ea;
					while ((NextEA - TempAddr) >= MD_DEFAULT_RETURN_ADDRESS_SIZE) {
						SMP_xref_t xrefs;
						for (bool ok = xrefs.SMP_first_from(TempAddr, XREF_DATA); ok; ok = xrefs.SMP_next_from()) {
							STARS_ea_t TargetAddr = xrefs.GetTo();
							if ((TargetAddr != 0) && (!xrefs.GetIscode())) {
								// Found a target, with its address in xrefs.to
								//  Is the target code?
								STARS_Segment_t *SegInfo = SMP_getseg(TargetAddr);
								if ((NULL != SegInfo) && (SegInfo->IsCodeSegment())) {
									bool NewTarget = CurrProg->InsertDataToCodeXref(TargetAddr);
									if (NewTarget)
										global_STARS_program->PrintDataToCodeXref(TempAddr, TargetAddr, 0);
								}
							}
						}
#ifdef __EA64__
						uint64_t DataValue = get_qword(TempAddr);
#else
						uint32_t DataValue = get_long(TempAddr);
#endif
						if (DataValue != 0) {
							// Is this a code address?
							STARS_ea_t PossibleCodeAddr = (STARS_ea_t) DataValue;
							STARS_Segment_t *SegInfo = SMP_getseg(PossibleCodeAddr);
							if ((NULL != SegInfo) && (SegInfo->IsCodeSegment())) {
								bool NewTarget = CurrProg->InsertDataToCodeXref(PossibleCodeAddr);
								if (NewTarget)
									global_STARS_program->PrintDataToCodeXref(TempAddr, PossibleCodeAddr, 0);
							}
						}
						TempAddr += MD_DEFAULT_RETURN_ADDRESS_SIZE;
					}

					// Move on to next data object
					ea = NextEA;
				}
				else {
					ea = nextaddr(ea);
				}
			} // end while (ea < seg->endEA)
		} // end if (seg->type == SEG_DATA ...)
		else if (seg->IsCodeSegment()) {
			if (seg->get_startEA() < TempLowestCodeAddress)
				TempLowestCodeAddress = seg->get_startEA();
			if (seg->get_endEA() > TempHighestCodeAddress)
				TempHighestCodeAddress = seg->get_endEA() - 1;
		} // end else if (seg->type === SEG_CODE)
		else {
#if SMP_DEBUG
			SMP_msg("Not processing segment of type <elided> SegName: %s from %lx to %lx\n",
				/*seg->type,*/ SegName, (unsigned long)seg->get_startEA(), (unsigned long)seg->get_endEA());
#endif
		}
	} // end for all segments
#if SMP_COUNT_MEMORY_ALLOCATIONS
	SMPGlobalVarCount += this->GlobalVarTable.size();
#endif

	// Update SMPProgram global limits
	if (LowestCodeAddress > TempLowestCodeAddress) {
		LowestCodeAddress = TempLowestCodeAddress;
		SMP_msg("INFO: New lowest code address: %x\n", LowestCodeAddress);
	}
	if (HighestCodeAddress < TempHighestCodeAddress) {
		HighestCodeAddress = TempHighestCodeAddress;
		SMP_msg("INFO: New highest code address: %x\n", HighestCodeAddress);
	}
	if (LowestGlobalVarAddress > TempLowestGlobalVarAddress) {
		LowestGlobalVarAddress = TempLowestGlobalVarAddress;
		SMP_msg("INFO: New lowest static data address: %x\n", LowestGlobalVarAddress);
	}
	if (HighestGlobalVarAddress < TempHighestGlobalVarAddress) {
		HighestGlobalVarAddress = TempHighestGlobalVarAddress;
		SMP_msg("INFO: New highest static data address: %x\n", HighestGlobalVarAddress);
	}

	return;
} // end of STARS_IDA_Program_t::InitStaticDataTable()
#endif

#if 0
// Find the direct and indexed accesses to offsets within each static data table entry.
//  Record the offset and kind of access (indexed or not) and conservatively mark the
//  field boundaries based on the unindexed accesses.
void STARS_IRDB_Program_t::ComputeGlobalFieldOffsets(struct GlobalVar &CurrGlobal) {
	SMP_xref_t xb;
	STARS_ea_t addr;
	std::size_t offset;
	bool DebugFlag = false;
	DebugFlag |= (0 == strcmp("spec_fd", CurrGlobal.name));
	for (addr = CurrGlobal.addr; addr < CurrGlobal.addr + CurrGlobal.size; ++addr) {
		bool Referenced = false;
		offset = addr - CurrGlobal.addr;
		pair<std::size_t, bool> TempOffset(offset, false); // false ==> No indexed accesses seen yet
		for (bool ok = xb.SMP_first_to(addr, XREF_ALL); ok; ok = xb.SMP_next_to()) {
			uchar XrefType = xb.GetType() & XREF_MASK;
			if (xb.GetIscode()) {
#if SMP_DEBUG_GLOBAL_GRANULARITY
				SMP_msg("WARNING: code xref to global data at %x\n", addr);
#endif
				;
			}
			else {
				STARS_ea_t FromAddr = xb.GetFrom();
				if ((XrefType == dr_O) || (XrefType == dr_W) || (XrefType == dr_R)) {
#if SMP_DEBUG_GLOBAL_GRANULARITY
					SMPInstr TempInstr(FromAddr);
					TempInstr.Analyze();
					SMP_msg("Data xref to global data %s at %x from code at %x %s\n",
						CurrGlobal.name, addr, FromAddr, TempInstr.GetDisasm());
#endif
					Referenced = true;
					TempOffset.second |= MDIsIndexedAccess(FromAddr, addr);
				}
				else {
#if SMP_DEBUG_GLOBAL_GRANULARITY
					SMP_msg("WARNING: Weird data xref type %d at %x\n", XrefType, FromAddr);
#endif
					;
				}
			}
		} // end for (bool ok = iterate through xrefs ...)
		if (Referenced) {
			CurrGlobal.FieldOffsets.insert(TempOffset);
		}
	} // end for all addrs in current global

	return;
} // end of STARS_IDA_Program_t::ComputeGlobalFieldOffsets()
// No need for IRDB version for fine-grained global static data, as all fine-grained info computed in ComputeGlobalFieldOffsets() will
//  be written to the annotations file, read by the IRDB builder process, and then read back in
//  the IRDB version of InitStaticDataTable().
#endif
